//---------------------------------------------------------------------------
// Copyright (c) Allan Petersen, 2000.
/*
 * This is a benchmark utility to view the differences between.
 * regular math routines and fastmath routines.

 * (c) Copyright 2000, Allan Petersen
 * ALL RIGHTS RESERVED
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice appear in all copies and that both the copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the name of Allan Petersen not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
 * AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
 * FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL ALLAN
 * PETERSEN BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
 * SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY
 * KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
 * LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
 * THIRD PARTIES, WHETHER OR NOT ALLAN PETERSEN HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Contact Allan Petersen at <support@allanpetersen.com> or visit
 * www.danbbs.dk/~allmette
 *
 */
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <fastmath.h>
#include "Unit1.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;

int sort_time (const void *a, const void *b);

//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
int sort_time (const void *a, const void *b)
{
    if (*(float*) a  > *(float*) b) return ( 1);
    if (*(float *)a  < *(float *)b) return (-1);
    if (*(float *)a == *(float *)b) return ( 0);
    return (0);
}
//---------------------------------------------------------------------------
// ---- Set of normal math functions and functions from the fastmath
// ---- library
//---------------------------------------------------------------------------
float __fastcall TForm1::BenchType (int  mode, float *Value, int *Index)
{
    switch (mode) {
        case 0:
            return(sqrt(*Value));
        case 1:
            return (_fm_sqrt(*Value));
        case 2:
            return (sin (*Value));
        case 3:
            return (_fm_sin (*Value));
        case 4:
            return (cos (*Value));
        case 5:
            return (_fm_cos (*Value));
        case 6:
            return (tan (*Value));
        case 7:
            return (_fm_tan (*Value));
        case 8:
            return (*Value / (*Index+1));
        case 9:
            return (*Value * (*Index+1));
        case 10:
            return (exp(*Value));
        case 11:
            return (_fm_exp(*Value));
        case 12:
            return (fabs(*Value));
        case 13:
            return (_fm_fabs(*Value));
        case 14:
            return (log(*Value));
        case 15:
            return (_fm_log(*Value));
        }
    return (0);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Bench (void)
{
#define     SAMPLES     100         // ---- Number of samples to collect
#define     N_RUNS      1000        // ---- Number of runs per sample
#define     N_MODES     16          // ---- Number of modes, N_MODES/2 = number of names

    char            str[255];
    // ---- names to write out    
    char           *StrModeNames[] = {"sqrt()",
                                      "sin()",
                                      "cos()",
                                      "tan()",
                                      "Div - Mult",
                                      "exp()",
                                      "fabs()",
                                      "log()"};
    int             ii,
                    jj,
                    xm,
                    cachesize;
    TListItem      *pItem;
    TListColumn    *pColumn;
    LARGE_INTEGER   Frequency,
                    TimerStart,
                    TimerStop;
    __int64         Latenzy;
    float          *atval,
                    t;
    float           UsedTime1,
                    UsedTime2,
                    UsedTime [N_MODES],
                    ausedtime[SAMPLES];
    float	       *pkcache;	// BIG ARRAY FOR OVERWRITING CACHE

    // ---- Making thread to run time critical to keep other programs from interupting
    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);

    ButtonBench->Enabled = false;

    ListView1->Items->Clear();
    ListView1->Update();


    // ---- Get frequncy of high performance timer
    QueryPerformanceFrequency(&Frequency);

    // ---- Measure the lantenzy in the performance counter
    QueryPerformanceCounter(&TimerStart);
    QueryPerformanceCounter(&TimerStop);
    // ---- Calculating time stamp latency
    Latenzy = TimerStop.QuadPart - TimerStart.QuadPart;

    atval = (float *) malloc (sizeof(float)*N_RUNS);

    cachesize = sizeof(float)*65536;
    pkcache   = (float *) malloc (cachesize);

    // ------------------------------------------------------------
    Caption = "Building Table";
    randomize();
    for (ii=0; ii<N_RUNS; ii++) {
        atval[ii] = (float)(rand() % 100) / ((float)(rand() % 100)+1.)+1.0;
        if ((ii % (N_RUNS/100)) == 0) {
            sprintf(str, "Building Table %.0f %%", ((float)ii/(float)N_RUNS)*100.);
            Caption = str;
            }
        }

    Caption = "Math Benchmark by AP (c) 2000";
    // ------------------------------------------------------------
    // ------------------------------------------------------------
    // ---- Looping modes
    for (xm=0; xm<N_MODES; xm++) {
        if ((xm % 2) == 0) {
            pItem = ListView1->Items->Add ();
            pItem->Caption = String (StrModeNames[xm/2]);
            pItem->Update ();
            }
        // ---- Looping samples
        for (jj=0; jj<SAMPLES; jj++) {
            if (CheckBoxKillCache->Checked) {
                memset(pkcache,0,cachesize);		// EMPTY DATA CACHE 64K TO BE SURE
                }
            QueryPerformanceCounter(&TimerStart);
            for (ii=0; ii<N_RUNS; ii++) {
                BenchType (xm, &atval[ii], &ii);
                }
            QueryPerformanceCounter(&TimerStop);
            ausedtime[jj] = (((float)TimerStop.QuadPart - (float)TimerStart.QuadPart)-(float)Latenzy) / (float)Frequency.QuadPart;
            }
        qsort ((void *)&ausedtime, SAMPLES, sizeof (float), sort_time);
        UsedTime[xm] = 0;
        // ---- rejecting first and last value
        for (jj=1; jj<SAMPLES-1; jj++) {
            UsedTime[xm] += ausedtime[jj];
            }
        // ---- calculating average
        UsedTime[xm] /= (float)SAMPLES-2;
        sprintf (str, "%.2fms", UsedTime[xm]*1000.);
        pItem->SubItems->Add(str);
        pItem->Update ();
        if (((xm+1) % 2) == 0 && xm > 0) {
            sprintf (str, "%.2f%%", ((float)(UsedTime[xm-1]-UsedTime[xm])/UsedTime[xm-1])*100.);
            pItem->SubItems->Add(str);
            pItem->Update ();
            }
        } // for xm...
    // ------------------------------------------------------------

    // ---- release any memory collected
    free (atval);
    free (pkcache);
    ButtonBench->Enabled = true;

    // ---- Making thread to run normal
    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::ButtonBenchClick(TObject *Sender)
{
    Bench ();
}
//---------------------------------------------------------------------------



